using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace H2Forge.Map.Stream
{
    public class MapStream
    {
        #region Instances
        private FileStream __fs;
        private BinaryReader __br;
        private BinaryWriter __bw;
        private string __filePath;
        private string __pluginPath;

        private Divisions.Header __header__;
        private Divisions.Index __index__;
        private Divisions.Tags __tags__;
        private Divisions.Sbsp __sbsp__;
        private Divisions.Strings __strings__;
        internal bool __sbspAndLtmpInfoIsFixed__ = false;
        #endregion

        #region Accessors
        private FileStream fs { get { return __fs; } set { __fs = value; } }
        private BinaryReader br { get { return __br; } set { __br = value; } }
        private BinaryWriter bw { get { return __bw; } set { __bw = value; } }
        public long Length { get { return fs.Length; } }
        public long Position { get { return fs.Position; } set { fs.Position = value; } }

        /// <summary>
        /// The Absolute File Path Of The Halo 2 Map
        /// </summary>
        public string FilePath { get { return __filePath; } }
        private string PluginPath { get { return __pluginPath; } set { __pluginPath = value; } }

        /// <summary>
        /// Halo 2 Map Header
        /// </summary>
        public Divisions.Header Header { get { return __header__; } set { __header__ = value; } }
        /// <summary>
        /// Halo 2 Map Index
        /// </summary>
        public Divisions.Index Index { get { return __index__; } set { __index__ = value; } }
        /// <summary>
        /// Halo 2 Map Tags
        /// </summary>
        public Divisions.Tags Tags { get { return __tags__; } set { __tags__ = value; } }
        /// <summary>
        /// Halo 2 Map Sbsp
        /// </summary>
        public Divisions.Sbsp Sbsp { get { return __sbsp__; } set { __sbsp__ = value; } }
        /// <summary>
        /// Halo 2 Map Strings
        /// </summary>
        public Divisions.Strings Strings { get { return __strings__; } set { __strings__ = value; } }
        /// <summary>
        /// Read Only Returns True If The Sbsp And Ltmp TagInfo Has Been Fixed
        /// </summary>
        public bool SbspAndLtmpInfoIsFixed { get { return __sbspAndLtmpInfoIsFixed__; } }
        #endregion

        #region Essentials
        internal int IndexOffset;
        internal int FileTableOffset;
        internal int SecondaryMagic;
        internal int TagInfoCount;
        internal int TagInfoIndexOffset;
        internal int MetaStart;
        internal Data.Types.Tag Scnr;
        internal Divisions.SbspInfo[] SbspInfoList;
        internal SortedDictionary<string, Data.Types.Tag> SbspTags;
        internal SortedDictionary<string, Divisions.LightMap> LightMapTags;
        //Tags - If sbsp or ltmp return from here
        #endregion

        #region New
        public MapStream(bool FixSbspAndLtmp)
        {
            OpenFileDialog OFD = new OpenFileDialog();
            OFD.Filter = "Halo 2 Map (*.map)|*.map";
            
            if (OFD.ShowDialog() == DialogResult.OK)
            {
                OpenMap(OFD.FileName, FixSbspAndLtmp);
            }
        }

        public MapStream(string filePath, bool FixSbspAndLtmp)
        {
            OpenMap(filePath, FixSbspAndLtmp);
        }

        private void OpenMap(string filePath, bool FixSbspAndLtmp)
        {
            if (File.Exists(filePath))
            {
                FileStream Map = new FileStream(filePath, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);
                fs = Map;
                br = new BinaryReader(fs);
                bw = new BinaryWriter(fs);

                __filePath = filePath;
                __sbspAndLtmpInfoIsFixed__ = FixSbspAndLtmp;

                Header = new Divisions.Header(this);
                Index = new Divisions.Index(this);
                Tags = new Divisions.Tags(this, FixSbspAndLtmp);
                Sbsp = new Divisions.Sbsp(this, FixSbspAndLtmp);
                Strings = new Divisions.Strings(this);
            }
            else
            {
                throw new Exception("File Does Not Exsist");
            }
        }
        #endregion

        #region Read
        public byte ReadByte()
        {
            return br.ReadByte();
        }

        public byte ReadByte(bool MovePosition)
        {
            byte value = br.ReadByte();
            if (!MovePosition) Position--;
            return value;
        }

        public byte[] ReadBytes(int Count)
        {
            return br.ReadBytes(Count);
        }

        public byte[] ReadBytes(int Count, bool MovePosition)
        {
            byte[] value = br.ReadBytes(Count);
            if (!MovePosition) Position -= Count;
            return value;
        }

        public char[] ReadChars(int Count)
        {
            char[] value = new char[Count];
            for (int i = 0; i < Count; i++)
            {
                value[i] = Convert.ToChar(br.ReadByte());
            }
            return value;
        }

        public char[] ReadChars(int Count, bool MovePosition)
        {
            char[] value = new char[Count];
            for (int i = 0; i < Count; i++) value[i] = Convert.ToChar(br.ReadByte());
            if (!MovePosition) Position -= Count;
            return value;
        }

        public float ReadFloat()
        {
            return br.ReadSingle();
        }

        public float ReadFloat(bool MovePosition)
        {
            float value = br.ReadSingle();
            if (!MovePosition) Position -= 2;
            return value;
        }

        public short ReadInt16()
        {
            return br.ReadInt16();
        }

        public short ReadInt16(bool MovePosition)
        {
            short value = br.ReadInt16();
            if (!MovePosition) Position -= 2;
            return value;
        }

        public int ReadInt32()
        {
            return br.ReadInt32();
        }

        public int ReadInt32(bool MovePosition)
        {
            int value = br.ReadInt32();
            if (!MovePosition) Position -= 4;
            return value;
        }

        public long ReadInt64()
        {
            return br.ReadInt64();
        }

        public long ReadInt64(bool MovePosition)
        {
            long value = br.ReadInt64();
            if (!MovePosition) Position -= 8;
            return value;
        }

        public Data.Types.Bitmask ReadBitmask(int ByteCount)
        {
            return new Data.Types.Bitmask(br.ReadBytes(ByteCount));
        }

        public Data.Types.Bitmask ReadBitmask(int ByteCount, bool MovePosition)
        {
            byte[] BitBytes = br.ReadBytes(ByteCount);

            if (!MovePosition) Position -= ByteCount;

            return new Data.Types.Bitmask(BitBytes);
        }

        public string ReadString(int CharCount, bool removeNull)
        {
            char[] chars = ReadChars(CharCount);
            return (removeNull ? new string(chars).Split('\0')[0] : new string(chars));
        }

        public string ReadString(int CharCount, bool removeNull, bool MovePosition)
        {
            char[] chars = ReadChars(CharCount, MovePosition);
            return (removeNull ? new string(chars).Split('\0')[0] : new string(chars));
        }

        public string ReadReverseString(int CharCount, bool removeNull)
        {
            char[] chars = ReadChars(CharCount);
            Array.Reverse(chars);
            return (removeNull ? new string(chars).Split('\0')[0] : new string(chars));
        }

        public string ReadReverseString(int CharCount, bool removeNull, bool MovePosition)
        {
            char[] chars = ReadChars(CharCount, MovePosition);
            Array.Reverse(chars);
            return (removeNull ? new string(chars).Split('\0')[0] : new string(chars));
        }

        public string ReadUnicodeString(int CharCount)
        {
            string str;

            Encoding dec = Encoding.Unicode;
            str = dec.GetString(br.ReadBytes(CharCount));
            return str.Split('\0')[0];
        }

        public string ReadUnicodeString(int CharCount, bool MovePosition)
        {
            string str;

            Encoding dec = Encoding.Unicode;
            str = dec.GetString(ReadBytes(CharCount, MovePosition));
            return str.Split('\0')[0];
        }

        public string ReadNullTerminatedString()
        {
            StringBuilder str = new StringBuilder();
            char tempChar;

            for (int x = 0; x < Length; x++)
            {
                tempChar = Convert.ToChar(br.ReadByte());

                if (tempChar == '\0') break;
                str.Insert(str.Length, tempChar);
            }
            return str.ToString();
        }

        public string ReadNullTerminatedString(bool MovePosition)
        {
            StringBuilder str = new StringBuilder();
            char tempChar;

            for (int x = 0; x < Length; x++)
            {
                tempChar = Convert.ToChar(br.ReadByte());

                if (tempChar == '\0')
                {
                    if (!MovePosition) Position -= (x - 1);
                    break;
                }
                str.Append(tempChar);
            }
            return str.ToString();
        }
        #endregion

        #region Read At
        public byte ReadByteAt(int Offset, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            byte value = br.ReadByte();
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public byte[] ReadBytesAt(int Offset, int Count, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            byte[] value = br.ReadBytes(Count);
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public char[] ReadCharsAt(int Offset, int Count, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            char[] value = new char[Count];
            for (int i = 0; i < Count; i++) value[i] = Convert.ToChar(br.ReadByte());
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public float ReadFloatAt(int Offset, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            float value = br.ReadSingle();
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public short ReadInt16At(int Offset, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            short value = br.ReadInt16();
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public int ReadInt32At(int Offset, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            int value = br.ReadInt32();
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public long ReadInt64At(int Offset, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            long value = br.ReadInt64();
            if (!MovePosition) Position = oldPosition;
            return value;
        }

        public Data.Types.Bitmask ReadBitmaskAt(int Offset, int ByteCount, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            byte[] BitBytes = br.ReadBytes(ByteCount);

            if (!MovePosition) Position = oldPosition;

            return new Data.Types.Bitmask(BitBytes);
        }

        public string ReadStringAt(int Offset, int CharCount, bool removeNull, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            char[] chars = new char[CharCount];
            for (int x = 0; x < CharCount; x++) chars[x] += Convert.ToChar(br.ReadByte());
            if (!MovePosition) Position = oldPosition;
            return (removeNull ? new string(chars).Split('\0')[0] : new string(chars));
        }

        public string ReadReverseStringAt(int Offset, int CharCount, bool removeNull, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            char[] chars = new char[CharCount];
            for (int x = 0; x < CharCount; x++) chars[x] += Convert.ToChar(br.ReadByte());
            if (!MovePosition) Position = oldPosition;
            Array.Reverse(chars);
            return (removeNull ? new string(chars).Split('\0')[0] : new string(chars));
        }

        public string ReadUnicodeStringAt(int Offset, int CharCount, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            string str;

            Encoding dec = Encoding.Unicode;
            str = dec.GetString(br.ReadBytes(CharCount));
            if (!MovePosition) Position = oldPosition;
            return str.Split('\0')[0];
        }

        public string ReadNullTerminatedStringAt(int Offset, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            StringBuilder str = new StringBuilder();
            char tempChar;

            for (int x = 0; x < Length; x++)
            {
                tempChar = Convert.ToChar(br.ReadByte());
                if (tempChar == '\0')
                {
                    if (!MovePosition) Position = oldPosition;
                    break;
                }
                str.Append(tempChar);

            }
            return str.ToString();
        }
        #endregion

        #region Write
        public void Write(byte value)
        {
            bw.Write(value);
        }

        public void Write(byte value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position--;
        }

        public void Write(byte[] value)
        {
            bw.Write(value);
        }

        public void Write(byte[] value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= value.Length;
        }

        public void Write(char[] value)
        {
            bw.Write(value);
        }

        public void Write(char[] value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= value.Length;
        }

        public void Write(float value)
        {
            bw.Write(value);
        }

        public void Write(float value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= 4;
        }

        public void Write(short value)
        {
            bw.Write(value);
        }

        public void Write(short value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= 2;
        }

        public void Write(int value)
        {
            bw.Write(value);
        }

        public void Write(int value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= 4;
        }

        public void Write(long value)
        {
            bw.Write(value);
        }

        public void Write(long value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= 8;
        }

        public void Write(Data.Types.Bitmask Bits)
        {
            byte[] newBytes = Data.Types.Bitmask.ConvertToByteArray(Bits.Bits);
            bw.Write(newBytes);
        }

        public void Write(Data.Types.Bitmask Bits, bool MovePosition)
        {
            byte[] newBytes = Data.Types.Bitmask.ConvertToByteArray(Bits.Bits);
            bw.Write(newBytes);
            if (!MovePosition) Position -= newBytes.Length;
        }

        public void Write(string value)
        {
            bw.Write(value);
        }

        public void Write(string value, bool MovePosition)
        {
            bw.Write(value);
            if (!MovePosition) Position -= value.Length;
        }

        public void WriteReverseString(string value)
        {
            char[] Chars = value.ToCharArray();
            Array.Reverse(Chars);
            bw.Write(value);
        }

        public void WriteReverseString(string value, bool MovePosition)
        {
            char[] Chars = value.ToCharArray();
            Array.Reverse(Chars);
            bw.Write(Chars);
            if (!MovePosition) Position -= Chars.Length;
        }
        #endregion

        #region Write At
        public void WriteAt(int Offset, byte value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, byte[] value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, char[] value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, float value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, short value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, int value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, long value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, Data.Types.Bitmask Bits, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            byte[] newBytes = Data.Types.Bitmask.ConvertToByteArray(Bits.Bits);
            bw.Write(newBytes);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteAt(int Offset, string value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            bw.Write(value);
            if (!MovePosition) Position = oldPosition;
        }

        public void WriteReverseStringAt(int Offset, string value, bool MovePosition)
        {
            long oldPosition = Position;
            Position = Offset;
            char[] Chars = value.ToCharArray();
            Array.Reverse(Chars);
            bw.Write(Chars);
            if (!MovePosition) Position = oldPosition;
        }
        #endregion

        #region Flush + Close
        public void FlushAndClose()
        {
            bw.Flush();
            bw.Close();
            br.Close();
        }

        public void Flush()
        {
            bw.Flush();
        }

        public void Close()
        {
            br.Close();
            bw.Close();
        }
        #endregion

        #region Operate
        /// <summary>
        /// Add Blank Chunks To UnNested Reflexive
        /// </summary>
        /// <param name="ID">ID of Tag</param>
        /// <param name="reflexive">Reflexive To Add Chunks To</param>
        public void AddBlankChunks(int InsertionCount, int ID, Map.Plugins.Items.ReflexiveTree reflexive, System.Windows.Forms.ProgressBar probar)
        {
            Map.Data.Divisions.Tags AllTags = new H2Forge.Map.Data.Divisions.Tags(Tags.ReadTags(true, probar));
            Map.Data.Types.Tag TagBeingAddedTo = AllTags.GetTag(ID);

            int TagStartOffset = TagBeingAddedTo.Offset;

            int OldChunkCount = ReadInt32At(TagStartOffset + reflexive.Offset, true);

            int InsertionOffset = ReadInt32() - SecondaryMagic + (reflexive.ChunkSize * OldChunkCount);
            int InsertionSize = reflexive.ChunkSize * InsertionCount;

            Map.Plugins.Reader PluginReader = new H2Forge.Map.Plugins.Reader(Application.StartupPath + "\\Plugins\\");
            Map.Plugins.Items.ReflexiveTree[] ReflexivesInClass;

            //Check + Fix All Reflexives
            foreach (KeyValuePair<string, SortedDictionary<string, Data.Types.Tag>> Class in AllTags.Items)
            {
                ReflexivesInClass = PluginReader.ReadAllReflexives(Class.Key);
                foreach (Data.Types.Tag Tag in Class.Value.Values)
                {
                    AlterPointers(1, 0, Tag.Offset, ReflexivesInClass, InsertionOffset, InsertionSize);
                }
            }

            WriteAt(TagStartOffset + reflexive.Offset, OldChunkCount + InsertionCount, false);

            //Add New Blank Data At EOF
            AddBytesToEOF(InsertionSize);

            //Shift Data Forward
            ShiftBytes(InsertionOffset, InsertionSize);
        }

        //private int CountBytes(Map.Plugins.Items.ReflexiveTree reflexive)
        //{ 
        //    int ByteCount = reflexive.ChunkSize ;
        //    int NextChunkCount;
        //    int NextOffset;

        //    for (int i = 0; i < reflexive.Reflexives.Length; i++)
        //    {
        //        NextChunkCount = ReadInt32At(
        //    }
        //}

        private void AlterPointers(int ChunkCount, int ChunkSize, int MetaStart, Map.Plugins.Items.ReflexiveTree[] reflexives, int InsertionOffset, int InsertionSize)
        {
            int AbsouluteOffset;
            int CurrentReflexiveChunkCount;
            int OldPointer;

            for (int chunk = 0; chunk < ChunkCount; chunk++)
            {
                for (int currentReflexive = 0; currentReflexive < reflexives.Length; currentReflexive++)
                {
                    AbsouluteOffset = MetaStart + (chunk * ChunkSize) + reflexives[currentReflexive].Offset;
                    CurrentReflexiveChunkCount = ReadInt32At(AbsouluteOffset, true);
                    OldPointer = ReadInt32();

                    AlterPointers(CurrentReflexiveChunkCount, reflexives[currentReflexive].ChunkSize, OldPointer - SecondaryMagic, reflexives[currentReflexive].Reflexives, InsertionOffset, InsertionSize);

                    if (OldPointer - SecondaryMagic > InsertionOffset) WriteAt(AbsouluteOffset + 4, OldPointer + InsertionSize, false);
                }
            }
        }

        private void AddBytesToEOF(int Count)
        {
            fs.SetLength(Length + Count);
        }

        /// <summary>
        ///  Moves All Bytes After Start Offset Forward, New Bytes Are Added For Overflow.
        /// </summary>
        /// <param name="startOffset"></param>
        /// <param name="offset"></param>
        public void ShiftBytes(int startOffset, int offset)
        {
            //Added The Required Number Of Bytes To Shift The Data
            AddBytesToEOF(offset);


            //Reads Bytes
            byte[] shiftBytes = ReadBytesAt(startOffset, (int)Length - startOffset - offset, false);
            //Moves Bytes Forward
            WriteAt(startOffset + offset, shiftBytes, false);

            //Sets "Empty Data" Before The Data That Was Shifted To 0
            WriteAt(startOffset, new byte[offset], false);
        }

        public void Resign()
        {
            long oldPosition = Position;
            int size = (int)Length - 2048;
            int times = size / 4;
            int result = 0;
            int x;

            fs.Seek(2048, SeekOrigin.Begin);
            for (x = 0; x < times; x++)
            {
                result ^= ReadInt32();
            }

            fs.Seek(720, SeekOrigin.Begin);
            Write(result);
            Header.Checksum = result;

            Position = oldPosition;
            System.Windows.Forms.MessageBox.Show(Header.InternalName + "'s New Signature Is " + result.ToString(), "Resigned");
        }
        #endregion
    }
}
